from .GoogleTokenSpan import GoogleTokenSpan
from .GoogleSentiment import GoogleSentiment


class GoogleSentence(GoogleTokenSpan):
	def __init__(self, dictionary, document):

		text = dictionary.pop('text')
		content = text.pop('content')
		begin = text.pop('begin_offset')
		end = begin + len(content)

		super().__init__(dictionary=dictionary, document=document, begin=begin, end=end)

		self._text = content
		sentiment = self._dictionary.pop('sentiment')
		self._sentiment = GoogleSentiment(score=sentiment.pop('score'), magnitude=sentiment.pop('magnitude'))
		self._index = None

	@property
	def text(self):
		return self._text

	@property
	def id(self):
		return self.document.id, 'sentence', self._index

	@property
	def sentiment(self):
		"""
		:rtype: GoogleSentiment
		"""
		return self._sentiment

	def __str__(self):
		return self.text

	def __repr__(self):
		return str(self)








import re

WORD = re.compile(r'\w+')


def tokenize(text):
	"""
	this function tokenizes text at a very high speed
	:param str text: text to be tokenized
	:rtype: list[str]
	"""
	words = WORD.findall(text)
	return words



import requests


class Repustate:
	def __init__(self, api_key=None, server='api.repustate.com', cache=None):
		"""
		:param str api_key:
		:param disk.Cache.Cache cache:
		"""
		self._api_key = api_key or '$APIKEY'
		self._server = server

		self._cache = cache
		if self._cache:
			self.get_entities = self._cache.make_cached(
				id='repustate_get_entities_function',
				function=self._get_entities,
				sub_directory='get_entities'
			)
			self.get_sentiment = self._cache.make_cached(
				id='repustate_get_sentiment_function',
				function=self._get_sentiment,
				sub_directory='get_sentiment'
			)
		else:
			self.get_entities = self._get_entities
			self.get_sentiment = self._get_sentiment

	@property
	def usage(self):
		response = requests.get(
			f'https://{self._server}/v4/{self._api_key}/usage.json',
			verify=False,
		)
		return response.json()

	def _get_entities(self, text, language='en'):
		response = requests.post(
			f'https://{self._server}/v4/{self._api_key}/entities.json',
			verify=False, data={'text': text, 'lang': language}
		)
		return response.json()

	def _get_sentiment(self, text, topics=None, language='en'):
		if isinstance(text, str):
			if topics is None:
				response = requests.post(
					f'https://{self._server}/v4/{self._api_key}/score.json',
					verify=False, data={'text': text, 'lang': language}
				)

			else:
				response = requests.post(
					f'https://{self._server}/v4/{self._api_key}/topic.json',
					verify=False, data={'text': text, 'topics': ','.join(topics) , 'lang': language}
				)

		elif isinstance(text, list):
			if topics is None:
				data = {f'text{i}': t for i, t in enumerate(text)}
				data['lang'] = language
				response = requests.post(
					f'https://{self._server}/v4/{self._api_key}/bulk-score.json',
					verify=False, data=data
				)
			else:
				response = None
		else:
			response = None

		return response.json()



from .NounChunk import NounChunk
from .Entity import Entity
from .BasicToken import BasicToken
from .EntityChunk import EntityChunk
from .EntityType import EntityType
from .Sentence import Sentence
from .remove_list_duplicates import remove_list_duplicates
from .join_punctuation import join_punctuation
from .split_into_sentences import split_into_sentences

from abstract import Graph
from abstract.graph_style import NodeStyle
import spacy


class Token(BasicToken):
	def __init__(self, obj, document):
		super().__init__(obj=obj, document=document)
		self._parents = None
		self._children = None

	def __add__(self, other):
		"""
		:param Token or Document other: adding two tokens (or several tokens) creates a Document
		:rtype: Document
		"""
		return Document(text=str(self).rstrip()+' '+str(other).lstrip(), nlp=self.nlp)

	@property
	def parents(self):
		"""
		:rtype: list[Token]
		"""
		if self._parents is None:
			self._parents = [Token(obj=x, document=self.document) for x in self.token.ancestors]
		return self._parents

	@property
	def children(self):
		"""
		:rtype: list[Token]
		"""
		if self._children is None:
			self._children = [Token(obj=x, document=self.document) for x in self.token.children]
		return self._children


class Document:
	def __init__(self, text=None, nlp=None, id=0, _doc=None):
		"""
		:type text: str
		"""
		spacy.prefer_gpu()
		if isinstance(text, list):
			text = ' '.join([str(x) for x in text])
		self._id = id

		if text is None and _doc is None:
			raise ValueError('Either text or _doc should be given!')
		elif text is None:
			self._doc = _doc
			self._text = _doc.text
			self._nlp = nlp
		elif _doc is None:
			self._doc = None
			self._text = str(text)
			self._nlp = nlp or spacy.load('en_core_web_sm')
		else:
			raise ValueError('Either text or _doc should be None!')

		self._doc = None
		self._tokens = None
		self._noun_chunks = None
		self._entity_chunks = None
		self._entities = None
		self._sentences = None
		self._sentences_method = None
		self._entity_graph = None
		self._syntax_graph = None

	@property
	def nlp(self):
		"""
		:rtype: spacy.tokens.doc.Doc or NoneType
		"""
		return self._nlp

	@property
	def text(self):
		return self._text

	def tokenize(self):
		self._tokens = remove_list_duplicates([Token(obj=x, document=self) for x in self.doc])
		self._entities = remove_list_duplicates([Entity(obj=span, document=self) for span in self.doc.ents])
		self._noun_chunks = remove_list_duplicates([
			NounChunk(obj=span, document=self) for span in self.doc.noun_chunks
		])
		self._entity_chunks = remove_list_duplicates([EntityChunk(entity=entity) for entity in self._entities])

	@property
	def id(self):
		return self._id

	@property
	def doc(self):
		"""
		:rtype: tokens.Doc
		"""
		if self._doc is None:
			self._doc = self._nlp(self._text)
		return self._doc

	def __repr__(self):
		return self.doc.__repr__()

	def __str__(self):
		return self.doc.__str__()

	def graph_str(self):
		if len(str(self)) > 50:
			return f'{self}'[0:47] + '...'
		else:
			return str(self)

	@property
	def sentences(self):

		"""
		uses spacy to split a Document class into a list of Sentence objects
		:rtype: list[Sentence]
		"""
		if self._sentences is None:
			self._sentences = self.get_sentences(method='spacy', return_type='Sentence')
		return self._sentences

	def get_sentences(self, method='spacy', return_type='Sentence'):

		"""
		:param str method: can be spacy, regex
		:param str return_type: if method is spacy, return_type can be str or Sentence
		:rtype: list[str] or list[Sentence]
		"""

		return_type = return_type.lower()

		sentence_method = method.lower()
		self._sentences_method = sentence_method

		if sentence_method == 'spacy':
			self._sentences = [Sentence(obj=x, document=self) for x in self.doc.sents]
			'''
			elif sentence_method == 'nltk':
				self._sentences = nltk.sent_tokenize(text=self.text)
			'''
		elif sentence_method == 'regex':
			self._sentences = split_into_sentences(text=self.text)

		else:
			raise ValueError(f'method:{method} is unknown!')

		if return_type == 'str':
			return [str(x) for x in self._sentences]

		elif return_type == 'document':
			return [Document(text=str(x)) for x in self._sentences]

		else:
			return self._sentences

	@property
	def tokens(self):

		"""
		:rtype: list[Token]
		"""
		if self._tokens is None:
			self.tokenize()
		return self._tokens

	@property
	def entities(self):
		"""
		:rtype: list[Entity]
		"""
		if self._entities is None:
			self.tokenize()
		return self._entities

	@property
	def noun_chunks(self):
		"""
		:rtype: list[NounChunk]
		"""
		if self._noun_chunks is None:
			self.tokenize()
		return self._noun_chunks

	@property
	def entity_chunks(self):
		"""
		:rtype: list[EntityChunk]
		"""
		if self._entity_chunks is None:
			self.tokenize()
		return self._entity_chunks

	@property
	def entity_graph(self):
		"""
		:rtype: Graph
		"""
		sentence_style = NodeStyle(text_size=7, shape='rect', shape_style='"rounded, filled"')
		noun_chunk_style = NodeStyle(
			fill_colour='lightpink', text_colour='black', text_size=8,
			shape='rect', shape_style='"rounded, filled"'
		)
		entity_style = NodeStyle(fill_colour='gold3', text_colour='black')
		entity_chunk_style = NodeStyle(fill_colour='gold', text_colour='black')

		if self._entity_graph is None:
			self._entity_graph = Graph(ordering=False)
			# add sentences to the graph
			if len(self.sentences) > 1:
				self._entity_graph.add_node(
					name=str(self.id), label=f'{self.graph_str()}\n({len(self.sentences)} sentences)',
					style=NodeStyle(text_size=7, shape='rect', shape_style='"rounded, filled"')
				)
				for i, sentence in enumerate(self.sentences):
					self._entity_graph.add_node(
						name=str(sentence.id), label=f'{sentence.graph_str()}\n(sentence {i+1})',
						style=sentence_style
					)
					self._entity_graph.connect(start=str(self.id), end=str(sentence.id))
			else:
				for sentence in self.sentences:
					self._entity_graph.add_node(
						name=str(sentence.id), label=f'{sentence.graph_str()}\n(one sentence)',
						style=sentence_style
					)

			# add tokens to the graph
			for token in self.tokens:
				self._entity_graph.add_node(
					name=str(token.id), label=f'{token.graph_str()}',
					style=NodeStyle(text_size=8, shape='rect', shape_style='"rounded, filled"')
				)

			# add noun_chunks to the graph
			for noun_chunk in self.noun_chunks:
				self._entity_graph.add_node(
					name=str(noun_chunk.id), label=f'{noun_chunk.graph_str()}',
					style=noun_chunk_style
				)

				# noun_chunk to its tokens
				for token in noun_chunk.tokens:
					self._entity_graph.connect(start=str(token.noun_chunk.id), end=str(token.id))

			# connect tokens to sentences (only the ones that don't have a noun chunk parent)
			for sentence in self.sentences:
				noun_chunks_connected_to_this_sentence = []
				for token in sentence.tokens:
					if not token.noun_chunk:
						self._entity_graph.connect(start=str(sentence.id), end=str(token.id))
					else:
						noun_chunk = token.noun_chunk
						if noun_chunk not in noun_chunks_connected_to_this_sentence:
							self._entity_graph.connect(start=str(sentence.id), end=str(noun_chunk.id))
							noun_chunks_connected_to_this_sentence.append(noun_chunk)

			# add entities to the graph
			for entity in self.entities:
				self._entity_graph.add_node(
					name=str(entity.id), label=f'{entity.graph_str()}',
					style=entity_style
				)
				# connect entity to its tokens
				for token in entity.tokens:
					self._entity_graph.connect(start=str(token.id), end=str(entity.id))

			# add entity_chunks to the graph
			for entity_chunk in self.entity_chunks:
				self._entity_graph.add_node(
					name=str(entity_chunk.id), label=f'{entity_chunk.graph_str()}',
					style=entity_chunk_style
				)

				# connect entity_chunk to its entities
				for entity in entity_chunk.entities:
					self._entity_graph.connect(start=str(entity.id), end=str(entity_chunk.id))

		return self._entity_graph

	@property
	def syntax_graph(self):
		"""
		:rtype: Graph
		"""
		sentence_style = NodeStyle(text_size=7, shape='rect', shape_style='"rounded, filled"')
		if self._syntax_graph is None:
			self._syntax_graph = Graph(ordering=False)

			# add sentences to the graph
			for sentence in self.sentences:
				self._syntax_graph.add_node(
					name=str(sentence.id),
					label=f'{sentence.graph_str()}\n(sentence)',
					style=sentence_style
				)

			for token in self.tokens:
				self._syntax_graph.add_node(name=str(token.id), label=token.graph_str())

			# connect tokens to sentences
			for sentence in self.sentences:
				for token in sentence.tokens:
					# only connect root tokens (tokens without parents) to the sentence
					if len(token.parents) < 1:
						self._syntax_graph.connect(start=str(sentence.id), end=str(token.id))

			for token in self.tokens:
				for child in token.children:
					dependency_label = str(child.dependency).replace('_', '\n')
					self._syntax_graph.connect(start=str(token.id), end=str(child.id), label=dependency_label)
		return self._syntax_graph

	def __add__(self, other):
		"""
		:type other: Document or Token
		:rtype: Document
		"""
		return self.__class__(text=str(self).rstrip()+' '+str(other).lstrip())

	def bind(self, chunk_type='entity_chunk'):
		"""
		:param str chunk_type: can be one of entity_chunk, noun_chunk, or entity
		:rtype: list[Token or EntityChunk or NounChunk or Entity]
		"""
		if chunk_type == 'entity_chunk':
			chunks = [x for x in self.tokens if x.entity_chunk is None] + self.entity_chunks
		elif chunk_type == 'entity':
			chunks = [x for x in self.tokens if x.entity is None] + self.entities
		elif chunk_type == 'noun_chunk':
			chunks = [x for x in self.tokens if x.noun_chunk is None] + self.noun_chunks
		else:
			raise ValueError(f'chunk_type "{chunk_type}" is not acceptable!')
		return sorted(chunks, key=lambda x: x.start)

	def mask_entity_types(self, entity_types, chunk_type='entity_chunk'):
		"""
		:param list[str] or list[EntityType] or str entity_types: entity types to be masked, it can be 'all'
		:param str chunk_type: can be one of entity_chunk, noun_chunk, or entity
		:rtype: str
		"""
		if entity_types == 'all':
			pass
		else:
			entity_types = [EntityType(name=x) if not isinstance(x, EntityType) else x for x in entity_types]
		chunks = self.bind(chunk_type=chunk_type)
		strings = []
		for chunk in chunks:
			string = str(chunk)
			if (chunk_type == 'entity_chunk' and isinstance(chunk, EntityChunk)) or \
				(chunk_type == 'noun_chunk' and isinstance(chunk, NounChunk)):
				for entity_type in sorted(chunk.entity_types):
					if entity_types == 'all':
						string = f'<{entity_type.long_name.upper()}>'
						break
					elif entity_type in entity_types:
						string = f'<{entity_type.long_name.upper()}>'
						break
			elif chunk_type == 'entity' and isinstance(chunk, Entity):
				if entity_types == 'all':
					string = f'<{chunk.entity_type.long_name.upper()}>'
				elif chunk.entity_type in entity_types:
					string = f'<{chunk.entity_type.long_name.upper()}>'
			strings.append(string)
		return ' '.join(join_punctuation(seq=[str(x) for x in strings]))



import warnings
import spacy
from spacy.lang.en import English
from .document import Document


class Linguist:
	def __init__(self, spacy_core='en_core_web_sm', prefer_gpu=None, wikipedia_api=None):
		if prefer_gpu is None:
			self._gpu = spacy.prefer_gpu()
		elif prefer_gpu:  # explicit True
			self._gpu = spacy.prefer_gpu()
			if not self._gpu:
				warnings.warn('GPU is not working!')
		else:
			self._gpu = False

		self._nlp = spacy.load(spacy_core)
		self._wikipedia = wikipedia_api

	@property
	def nlp(self):
		"""
		:rtype: English
		"""
		return self._nlp

	def create_document(self, text):
		doc = self.nlp(text)
		return Document(_doc=doc, nlp=self.nlp)

	def analyze(self, text):
		return self.create_document(text=text)





from editdistance import eval as _get_levenshtein_distance


def get_levenshtein_distance(s1, s2, case_sensitive=True):
	if not case_sensitive:
		s1 = s1.lower()
		s2 = s2.lower()
	return _get_levenshtein_distance(s1, s2)


def get_levenshtein_similarity(s1, s2, case_sensitive=True):
	return 1-get_levenshtein_distance(s1, s2, case_sensitive=case_sensitive)/max(len(s1), len(s2))




from slytherin import trim


class GoogleDocumentObject:
	def __init__(self, dictionary, document, begin, end):
		"""
		:param dict dictionary: any object from a document
		:type document: .GoogleDocument.GoogleDocument
		"""
		self._dictionary = dictionary
		self._document = document
		self._begin = begin
		self._end = end

	@property
	def begin(self):
		return self._begin

	@property
	def end(self):
		return self._end

	def __lt__(self, other):
		"""
		:type other: GoogleToken
		:rtype: bool
		"""
		if self.begin < other.begin and self.end < other.end:
			return True
		elif self.begin >= other.begin and self.end >= other.end:
			return False
		else:
			raise ValueError(f'Token "{self}" and "{other}" overlap!')

	def __gt__(self, other):
		return other.__lt__(self)

	def __le__(self, other):
		return not self.__gt__(other)

	def __ge__(self, other):
		return not self.__lt__(other)

	@property
	def document(self):
		"""
		:rtype: .GoogleDocument.GoogleDocument
		"""
		return self._document

	def __str__(self):
		return self._dictionary.__str__()

	def graph_str(self):
		return trim(string=str(self), max_length=50, cut_from='middle')

	@property
	def id(self):
		return None

	def __eq__(self, other):
		return self.id == other.id

	def __ne__(self, other):
		return self.id != other.id

	@property
	def cloud(self):
		"""
		:rtype: .GoogleCloud.GoogleCloud
		"""
		return self.document.cloud



from setuptools import setup, find_packages
import spacy
# manually make sure en_core_web_sm is installed
import os
try:
	import spacy
	spacy.prefer_gpu()
	
	try:
		nlp = spacy.load('en_core_web_sm')
	except:
		os.system('python -m spacy download en')
except:
	os.system('pip install spacy')
	os.system('python -m spacy download en')


def readme():
	with open('./README.md') as f:
		return f.read()


setup(
	name='linguistics',
	version='2019.12.29.1',
	license='MIT',
	url='https://github.com/idin/linguistics',
	author='Idin',
	author_email='py@idin.ca',
	description='Python library for natural language processing',
	long_description=readme(),
	long_description_content_type='text/markdown',
	classifiers=[
		'Development Status :: 3 - Alpha',
		'Intended Audience :: Developers',
		'Programming Language :: Python :: 3 :: Only',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: 3.7',
		'Topic :: Software Development :: Libraries :: Python Modules'
	],
	packages=find_packages(exclude=["jupyter_tests", ".idea", ".git"]),
	install_requires=[
		'abstract', 'spacy', 'requests', 'jellyfish', 'editdistance', 'chronometry', 'slytherin', 'numpy', 'pandas'
	],
	python_requires='~=3.6',
	zip_safe=False
)


from chronometry.progress import ProgressBar

from .get_string_similarity import get_string_similarity
from .get_sentence_similarity import get_sentence_similarity


def get_similarity(
		s1, s2, method='jaro_winkler', similarity_function=None,
		case_sensitivity=1.0, first_char_weight=0.0, first_word_weight=0.0
):
	"""
	:type s1: str
	:type s2: str
	:type treat_as_sentence: bool
	:param str method: can be one of 'jaro_winkler', 'levenshtein', 'sentence_jaro_winkler', 'sentence_levenshtein'
	:type case_sensitivity: float
	:type first_char_weight: float
	:type first_word_weight: float
	:rtype: float
	"""
	if similarity_function is not None:
		return similarity_function(s1, s2)
	else:
		if method.startswith('sentence_'):
			treat_as_sentence = True
			method = method.split('_', 1)[1]
		else:
			treat_as_sentence = False

	if treat_as_sentence:
		return get_sentence_similarity(
			words1=s1, words2=s2, method=method, first_char_weight=first_char_weight, first_word_weight=first_word_weight,
			case_sensitivity=case_sensitivity
		)

	else:
		return get_string_similarity(
			s1=s1, s2=s2, method=method, first_char_weight=first_char_weight, case_sensitivity=case_sensitivity
		)


def get_similarities(
		string, strings, method='jaro_winkler', similarity_function=None,
		case_sensitivity=1.0,first_char_weight=0.0, first_word_weight=0.0, echo=0
):
	"""
	:type string: str
	:type strings: list[str]
	:type treat_as_sentence: bool
	:param str method: can be one of 'jaro_winkler', 'levenshtein', 'sentence_jaro_winkler', 'sentence_levenshtein'
	:type case_sensitivity: float
	:type first_char_weight: float
	:type first_word_weight: float
	:rtype:
	"""
	echo = max(0, echo)
	string = str(string)
	text = string + ' ? '
	return list(ProgressBar.map(
		function=lambda x: get_similarity(
			s1=string, s2=x, method=method, similarity_function=similarity_function,
			first_char_weight=first_char_weight, case_sensitivity=case_sensitivity, first_word_weight=first_word_weight
		),
		iterable=strings, iterable_text=strings, text=text, echo=echo
	))





from chronometry.progress import ProgressBar
import pandas as pd

from .get_similarity import get_similarities

def find_most_similar_for_one_string(
		string, candidates, candidate_ids=None, string_id=None, method='jaro_winkler', similarity_function=None,
		case_sensitivity=1.0, first_char_weight=0.0, first_word_weight=0.0, num_results=1, echo=0
):
	"""
	:type string: str
	:type candidates: list[str]
	:type candidate_ids: list or NoneType
	:type string_id: list or NoneType
	:param str method: can be one of 'jaro_winkler', 'levenshtein', 'sentence_jaro_winkler', 'sentence_levenshtein'
	:type treat_as_sentence: bool
	:type case_sensitivity: float
	:type first_char_weight: float
	:type first_word_weight: float
	:type num_results: int
	:type echo: int
	:rtype: pd.DataFrame
	"""
	echo = max(0, echo)
	num_results = max(1, min(num_results, len(candidates)))
	similarities = get_similarities(
		string=string, strings=candidates, method=method, similarity_function=similarity_function,
		case_sensitivity=case_sensitivity, first_char_weight=first_char_weight, first_word_weight=first_word_weight,
		echo=echo
	)
	if candidate_ids is None:
		candidate_ids = list(range(len(candidates)))
	df = pd.DataFrame({'string': str(string), 'candidate_id': candidate_ids, 'candidate': candidates, 'similarity': similarities})
	if string_id is not None:
		df['string_id'] = string_id


	df = df.sort_values(by='similarity', ascending=False)
	df = df[~df['candidate_id'].duplicated()].head(num_results)
	df['similarity_rank'] = range(1, df.shape[0] + 1)
	return df


def find_most_similar(
	strings, candidates, candidate_ids=None, string_ids=None, method='jaro_winkler', similarity_function=None,
	case_sensitivity=1.0, first_char_weight=0.0, first_word_weight=0.0, num_results=1, echo=0
):
	"""
	:type strings: str or list[str]
	:type candidates: list[str]
	:type candidate_ids: list or NoneType
	:type string_ids: list or NoneType
	:param str method: can be one of 'jaro_winkler', 'levenshtein', 'sentence_jaro_winkler', 'sentence_levenshtein'
	:type treat_as_sentence: bool
	:type first_char_weight: float
	:type first_word_weight: float
	:type case_sensitivity: float
	:rtype: pd.DataFrame
	"""
	echo = max(0, echo)
	if string_ids is None:
		string_ids = range(len(strings))

	return pd.concat(list(ProgressBar.map(
		function=lambda x: find_most_similar_for_one_string(
			string=x[0], string_id=x[1], candidates=candidates, candidate_ids=candidate_ids,
			method=method, similarity_function=similarity_function, case_sensitivity=case_sensitivity,
			first_char_weight=first_char_weight, first_word_weight=first_word_weight, num_results=num_results,
			echo=echo-1
		),
		iterable=list(zip(strings, string_ids)), iterable_text=strings, echo=echo
	))).reset_index(drop=True)[['string_id', 'string', 'candidate_id', 'candidate', 'similarity_rank', 'similarity']]